from astropy.io import fits
from astropy import wcs
import mpi4py.MPI as MPI
import galsim
import os
import yaml

from SingleEpochImage import SingleEpochImage
from InjectionCatalog import InjectionCatalog
from Catalog.C5_SimCat import SimCat
from Grid import RectGrid
from config import parse_args

class InjectionPipeline(object):

    def __init__(self):
        # Load configuration
        args = parse_args()
        if args.config_dir is None:
            args.config_dir = ''
        args.config_dir = os.path.abspath(args.config_dir)
        args.config_file = os.path.join(args.config_dir, args.config_file)
        with open(args.config_file, "r") as stream:
            try:
                self.config = yaml.safe_load(stream)
                for key, value in self.config.items():
                    print (key + " : " + str(value))
            except yaml.YAMLError as exc:
                print(exc)
        # Read the input image paths
        with open(self.config["input_img_list"]) as input_list:
            self.img_list = [line.rstrip() for line in input_list]

        # Prepare the output directory
        if not os.path.exists(self.config["output_img_dir"]):
            try:
                os.makedirs(self.config["output_img_dir"])
            except OSError:
                pass
        
        self.output_dir = os.path.join(self.config["output_img_dir"], self.config["run_name"])
        if not os.path.exists(self.output_dir):
            try:
                os.makedirs(self.output_dir)
            except OSError:
                pass

    def run_injection_img_list(self, Catalog):
        if self.config["use_mpi"]:
            comm = MPI.COMM_WORLD
            ind_thread = comm.Get_rank()
            num_thread = comm.Get_size()
        
        for i in range(len(self.img_list)):
            if self.config["use_mpi"]:
                if i % num_thread != ind_thread:
                    continue
            img_path = self.img_list[i]
            # Prepare output image path
            img_name = img_path.split('/')[-1].split('.')[0]
            output_name = img_name + '_injected.fits'
            output_path = os.path.join(self.output_dir, output_name)

            image = SingleEpochImage(config=self.config, input_img_path=img_path, output_img_path=output_path)
            inject_cat = InjectionCatalog(image=image)
            inject_cat.generate_positions(config=self.config)
            print("number of galaxies to be injected: %d"%(inject_cat.nobjects))
            input_cat = Catalog(config=self.config, chip=image.chip, nobjects=inject_cat.nobjects)
            image.inject_objects(pos=inject_cat.pos, cat=input_cat)
            image.save_injected_img()
    
if __name__ == "__main__":
    pipeline = InjectionPipeline()
    pipeline.run_injection_img_list(Catalog=SimCat)